# -*- coding: utf-8 -*-
#
# Copyright © 2009-2021 CEA
# Pierre Raybaut
# Licensed under the terms of the CECILL License
# (see guidata/__init__.py for details)

"""
widgets
=======

The ``guidata.widgets`` package provides useful Qt-based widgets.

.. automodule:: guidata.widgets.console
   :members:

.. automodule:: guidata.widgets.codeeditor
   :members:

.. automodule:: guidata.widgets.arrayeditor
   :members:

.. automodule:: guidata.widgets.collectionseditor
   :members:

.. automodule:: guidata.widgets.dataframeeditor
   :members:

.. automodule:: guidata.widgets.texteditor
   :members:

.. automodule:: guidata.widgets.objecteditor
   :members:
"""
